<?php

class GILSGitlab{
    private $curl = false;
    private $redis = false;
    
    /** Check whether a project falls within the permitted list
     * 
     * @arg proj_path - the URL safe project path
     * 
     * @return is_allowed (boolean)
     */
    function enforceProjectFilters($proj_path){
        global $config;
        
        if (!$config->enforce_explicit_limits){
            // Config option is turned off, so permit access
            return true;
        }
        
        // Otherwise, check whether the value exists in the config array
        if (in_array($proj_path, $config->permitted_projects)){
            return true;
        }
        
        // Finally, check if it exists in $permitted_groups (websites/Gitlab-Issue-Listing-Script#40)
        
        $proj_parts = explode("/",$proj_path);
        $group = $proj_parts[0];
        if (in_array($group, $config->permitted_groups)){
            return true;
        }        
        
        // Deny access
        return false;
    }
    
    function genCSS($proj_path, $type){
        global $config;
        
        if (!$this->enforceProjectFilters($proj_path)){
            http_response_code(404);
            echo "Commit Not found";
            die;
        }
        
        header("Content-Type: text/css");
        
        
        // Labels CSS is the only type currently supported
        if ($type == "labels"){
            $label_css = array();
            foreach($this->getProjectLabels($proj_path) as $label){
                $label_css[] = GILSUtils::buildLabelCSS($label);
            }     
            echo implode("\n",$label_css);
        }
        die;
    }
    
    function getCommit($proj_path, $commit_h, $enforce_filters = true){
        global $config;
        
        if ($enforce_filters && !$this->enforceProjectFilters($proj_path)){
            http_response_code(404);
            echo "Commit Not found";
            die;
        }        
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/repository/commits/" . $commit_h;
        
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        $commit = json_decode($r['result']);
        
        if (is_object($commit) && str_contains($commit->message, "Not Found")){
            http_response_code(404);
            echo "Not found";
            die;
        }
        
        $commit->title = GILSUtils::replaceCommitRefsEtc($commit->title, $proj_path);
        $commit->message = GILSUtils::replaceCommitRefsEtc($commit->message, $proj_path);
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/repository/commits/" . $commit_h . "/diff";
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        $commit_diff = json_decode($r['result']);
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/repository/commits/" . $commit_h . "/comments";
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        $comments = json_decode($r['result']);
        
        
        $comments_arr = array();
        foreach ($comments as $comment){
            $comment->note = GILSUtils::replaceCommitRefsEtc($comment->note, $proj_path);
            $comments_arr[] = $comment;
        }
        
        $o = array(
            "commit" => $commit,
            "diff" => $commit_diff,
            "comments" => $comments_arr,
            "signature" => $this->getCommitSignature($proj_path, $commit_h),
            "project" => $this->getProject($proj_path, $enforce_filters)
        );
        
        return $o;
    }
    
    
    function getCommitSignature($proj_path, $commit_h){
        global $config;
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/repository/commits/" . $commit_h . "/signature";
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        $sig = json_decode($r['result']);        
        
        $signature = array(
                "is_signed" => false
            );
        // 404 means it wasn't signed
        if (is_object($sig) && !property_exists($sig, "message") && $sig->signature_type == "PGP"){
            $signature["is_signed"] = true;
            $signature["status"] = $sig->verification_status;
            $signature["signed_by_key"] = $sig->gpg_key_primary_keyid;
            $signature["signed_by_user"] = $sig->gpg_key_user_name;
        }
        
        return $signature;
    }
    
    function getCommits($proj_path){
        global $config;
        
        if (!$this->enforceProjectFilters($proj_path)){
            http_response_code(404);
            echo "Commit List Not found";
            die;
        }        
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/repository/commits/";
        
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        
        $commits = json_decode($r['result']);
        $commit_arr = array();
        
        if (!is_array($commits) && property_exists($commits, "message") && str_contains($commits->message, "Not Found")){
            http_response_code(404);
            echo "Not found";
            die;
        }
        
        $x = 0;
        while (true){
            if (!$commits){
                break;
            }
            foreach ($commits as $commit) {
                $commit->signature = $this->getCommitSignature($proj_path, $commit->id);
                $commit->title = GILSUtils::replaceCommitRefsEtc($commit->title, $proj_path);
                $commit->message = GILSUtils::replaceCommitRefsEtc($commit->message, $proj_path);
                $commit_arr[] = $commit;
            }
            
            if (array_key_exists("next", $r['links'])){
                $b = $this->callAPI("GET", GILSUtils::forceHTTPS($r['links']["next"]["link"]), false, $config->do_auth);
                $r = $b;
                $commits = json_decode($r['result']);
            }else{
                break;
            }
            $x++;
        }
        
        $o = array(
            "commits" => $commit_arr,
            "project" => $this->getProject($proj_path)
        );

        return $o;        
    }
    
    
    function getIssuesForLabel($proj_path, $label){
        global $config;
        if (!$this->enforceProjectFilters($proj_path)){
            http_response_code(404);
            echo "Commit List Not found";
            die;
        }
        
        // Get the label so we can retrieve it's name (we'll usually be called by ID as of websites/Gitlab-Issue-Listing-Script#46
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/labels/" . rawurlencode($label);
        $l = $this->callAPI("GET", $url, false, $config->do_auth);

        if ($l['result'] == '{"message":"404 Label Not Found"}'){
            http_response_code(404);
            echo "Not found";
            die;
        }
        $label_obj = json_decode($l["result"]);
        
        // Retrieve associated issues
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/issues?sort=asc&order_by=created_at&labels=" . urlencode($label_obj->name);
        //echo $url;
        $r = $this->callAPI("GET", $url, false, $config->do_auth);

        if ($r['result'] == '{"message":"404 Project Not Found"}'){
            http_response_code(404);
            echo "Not found";
            die;
        }
        
        $issues = json_decode($r['result']);
        $issues_arr = array();
        $x = 0;
        while (true){
            if (!$issues){
                break;
            }
            foreach ($issues as $issue) {

                $dt = new DateTime($issue->created_at);
                $d = $dt->format('M-y');
                
                $p = array(
                    "title" => $issue->title,
                    "description" => GILSUtils::replaceCommitRefsEtc($issue->description, $proj_path),
                    "state" => $issue->state,
                    "created" => $issue->created_at,
                    "closed" => $issue->closed_at,
                    "closed_by" => $issue->closed_by,
                    "labels" => $issue->labels,
                    "milestone" => $issue->milestone,
                    "assignees" => $issue->assignees,
                    "author" => $issue->author,
                    "type" => $issue->issue_type,
                    "assignee" => $issue->assignee,
                    "gitlab_url" => $issue->web_url,
                    "gitlab_reference" =>  $issue->references->full,
                    "short_reference" =>  $issue->references->short
                );
                
                if (!array_key_exists($d, $issues_arr)){
                    $issues_arr[$d] = array();
                }
                
                $issues_arr[$d][] = $p;
                $x++;
            }
            
            if (array_key_exists("next", $r['links'])){
                $b = $this->callAPI("GET", GILSUtils::forceHTTPS($r['links']["next"]["link"]), false, $config->do_auth);
                $r = $b;
                $issues = json_decode($r['result']);
            }else{
                break;
            }
            
        }
    
        //$o["issues"] = $issues_arr;
        $r = array (
            "project" => $this->getProject($proj_path),
            "label" => $label_obj->name,
            "label_obj" => $label_obj,
            "num" => 0,
            "start" => "",
            "end" => "",
            "issues" => $issues_arr,
        );
        
        
        if ($x > 0){
            $last = end($issues_arr);
            $r["start"] = reset($issues_arr)[0]["created"];
            $r["end"] = end($last)["created"];
            $r["num"] = $x;
        }
                
        return $r;
    }
    
    
    function getContainers($proj_path){
        global $config;
        
        if (!$this->enforceProjectFilters($proj_path)){
            http_response_code(404);
            echo "Containers Not found";
            die;
        }        
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/registry/repositories/?tags=true";
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        
        $containers = json_decode($r['result']);    
        
        if (!$containers || (is_object($containers) && $containers->message)){
            $containers = array();
        }
        
        $o = array(
            "containers" => $containers,
            "project" => $this->getProject($proj_path)
        );

        return $o;           
    }

    function getMileStone($proj_path, $mid){
        global $config;
        
        if (!$this->enforceProjectFilters($proj_path)){
            http_response_code(404);
            echo "Milestone Not found";
            die;
        }        
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/milestones/" . $mid;
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        $milestone = json_decode($r['result']);
        
        
        $o = array(
                "id" => $milestone->id,
                "title" => $milestone->title,
                "project" => $proj_path,
                "description" => $milestone->description,
                "state" => $milestone->state,
                "created" => $milestone->created_at,
                "updated" => $milestone->updated_at,
                "due" => $milestone->due_date,
                "started" => $milestone->start_date,
                "web_url" => $milestone->web_url,
                "project" => $this->getProject($proj_path),
                "labels" => $this->getProjectLabels($proj_path)
             );
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/milestones/" . $mid . "/issues";
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        
        $issues = json_decode($r['result']);
        $issues_arr = array();
        
        $x = 0;
        while (true){
            if (!$issues){
                break;
            }
            foreach ($issues as $issue) {

                $p = array(
                    "title" => $issue->title,
                    "description" => GILSUtils::replaceCommitRefsEtc($issue->description, $proj_path),
                    "state" => $issue->state,
                    "created" => $issue->created_at,
                    "closed" => $issue->closed_at,
                    "closed_by" => $issue->closed_by,
                    "labels" => $issue->labels,
                    "milestone" => $issue->milestone,
                    "assignees" => $issue->assignees,
                    "author" => $issue->author,
                    "type" => $issue->issue_type,
                    "assignee" => $issue->assignee,
                    "gitlab_url" => $issue->web_url,
                    "gitlab_reference" =>  $proj_path . "#" . $issue->iid,
                    "short_reference" =>  "#" . $issue->iid
                );
                
                $issues_arr[] = $p;
            }
            
            if (array_key_exists("next", $r['links'])){
                $b = $this->callAPI("GET", GILSUtils::forceHTTPS($r['links']["next"]["link"]), false, $config->do_auth);
                $r = $b;
                $issues = json_decode($r['result']);
            }else{
                break;
            }
            $x++;
        }
        $o['issues'] = $issues_arr;
        return $o;
    }
    
    function getRelease($proj_path, $rid){
        global $config;
        
        if (!$this->enforceProjectFilters($proj_path)){
            http_response_code(404);
            echo "Release Not found";
            die;
        }
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/releases/" . $rid;
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        $release = json_decode($r['result']);
        $o = array(
                "release" => $release,
                "project" => $this->getProject($proj_path)
             );
        
        $o['release']->description = GILSUtils::replaceCommitRefsEtc($o['release']->description, $proj_path);
        
        return $o;
    }
    
    
    function getProjectList(){
        global $config;

        $url = $config->server . "/api/v4/projects?sort=asc&order_by=name";
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        
        $projects = json_decode($r['result']);
        $project_arr = array();
        $x = 0;
        while (true){
            if (!$projects){
                break;
            }

            foreach ($projects as $project) {

                if (!$this->enforceProjectFilters($project->path_with_namespace)){
                    continue;
                }
                
                if (property_exists($project, "namespace") && 
                    is_object($project->namespace) && 
                    in_array($project->namespace->path, $config->excluded_groups_from_homepage)){
                    
                    continue;
                }
                
                $p = array(
                    "name" => $project->path,
                    "namespaced_name" => $project->name_with_namespace,
                    "project_key" => $project->path_with_namespace,
                    "gitlab_url" => $project->web_url,
                    "description" => $project->description,
                    "avatar" => $project->avatar_url,
                    "group" => false
                );
                
                if (property_exists($project, "namespace") && 
                    is_object($project->namespace)){
                    $p["group"] = $project->namespace->name;
                }                
                                    
                
                $project_arr[] = $p;
            }
            
            if (array_key_exists("next", $r['links'])){
                $b = $this->callAPI("GET", GILSUtils::forceHTTPS($r['links']["next"]["link"], false, $config->do_auth));
                $r = $b;
                $projects = json_decode($r['result']);
            }else{
                break;
            }
            $x++;
        }
        return $project_arr;
    }
    

    function getProjectAndIssues($proj_path){
        //https://docs.gitlab.com/ee/api/projects.html#get-single-project
        
        global $config;
        
        if (!$this->enforceProjectFilters($proj_path)){
            http_response_code(404);
            echo "ProjectP Not found";
            die;
        }        
        
        $o = array(
            "project" => $this->getProject($proj_path),
            "labels" => $this->getProjectLabels($proj_path),
        );
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/issues?sort=asc&order_by=created_at";
        //echo $url;
        $r = $this->callAPI("GET", $url, false, $config->do_auth);

        if ($r['result'] == '{"message":"404 Project Not Found"}'){
            http_response_code(404);
            echo "Not found";
            die;
        }
        
        $issues = json_decode($r['result']);
        $issues_arr = array();
        $x = 0;
        while (true){
            if (!$issues){
                break;
            }
            foreach ($issues as $issue) {

                $p = array(
                    "title" => $issue->title,
                    "description" => GILSUtils::replaceCommitRefsEtc($issue->description, $proj_path),
                    "state" => $issue->state,
                    "created" => $issue->created_at,
                    "closed" => $issue->closed_at,
                    "closed_by" => $issue->closed_by,
                    "labels" => $issue->labels,
                    "milestone" => $issue->milestone,
                    "assignees" => $issue->assignees,
                    "author" => $issue->author,
                    "type" => $issue->issue_type,
                    "assignee" => $issue->assignee,
                    "gitlab_url" => $issue->web_url,
                    "gitlab_reference" =>  $issue->references->full,
                    "short_reference" =>  $issue->references->short
                );
                
                $issues_arr[] = $p;
            }
            
            if (array_key_exists("next", $r['links'])){
                $b = $this->callAPI("GET", GILSUtils::forceHTTPS($r['links']["next"]["link"]), false, $config->do_auth);
                $r = $b;
                $issues = json_decode($r['result']);
            }else{
                break;
            }
            $x++;
        }
    
        $o["issues"] = $issues_arr;
        
        // Check whether the project has a wiki
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/wikis";
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        
        if ($r['result'] != '{"message":"403 Forbidden"}'){
            $o['has_wiki'] = true;
        }else{
            $o['has_wiki'] = false;
        }
        
        
        $c = $this->getContainers($proj_path);
        $o['has_containers'] = (count($c['containers']) > 0)? true : false;
        
        // Get milestones
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/milestones";
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        $o["milestones"] = json_decode($r['result']);
        if (is_object($o["milestones"])){
            $o["milestones"] = false;
        }
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/releases";
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        $o["releases"] = json_decode($r['result']);
        if (is_object($o["releases"])){
            $o["releases"] = false;
        }
        
        return $o;
    }
    
    function getProjectWikiPages($proj_path){
        global $config;
        
        if (!$this->enforceProjectFilters($proj_path)){
            http_response_code(404);
            echo "Wiki Not found";
            die;
        }        
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/wikis";
        
        $r = $this->callAPI("GET", $url, false, $config->do_auth);

        if ($r['result'] == '{"message":"403 Forbidden"}'){
            http_response_code(404);
            echo "No exist";            
            die();
        }
        
        // Otherwise, crack on
        $pages = json_decode($r['result']);
        $pages_arr = array();
        while (true){
            if (!$pages){
                break;
            }
            foreach ($pages as $page) {               
                $pages_arr[] = $page;
            }
            
            if (array_key_exists("next", $r['links'])){
                $b = $this->callAPI("GET", GILSUtils::forceHTTPS($r['links']["next"]["link"]), false, $config->do_auth);
                $r = $b;
                $pages = json_decode($r['result']);
            }else{
                break;
            }
            $x++;
        }

        $o = array(
            "pages" => $pages_arr,
            "project" => $this->getProject($proj_path)
            );   
        return $o;
    }
    
    function getWikiPage($proj_path, $slug){
        global $config;
        
        if (!$this->enforceProjectFilters($proj_path)){
            http_response_code(404);
            echo "WikiPage Not found";
            die;
        }        
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/wikis/" . $slug;
        $r = $this->callAPI("GET", $url, false, $config->do_auth);

        if ($r['result'] == '{"message":"403 Forbidden"}'){
            echo "No exist";
            http_response_code(404);
            die();
        }
        // Otherwise, crack on
        $page = json_decode($r['result']);
        
        $page->gitlab_url = $config->server . "/" . $proj_path . "/-/wikis/" . $page->slug;

        
        
        $o = array(
            "page" => $page,
            "project" => $this->getProject($proj_path)
        );    
        return $o;
    
    }
    
    function getImage($path){
        global $config;
        $url = $config->server . $path;
        $r = $this->callAPI("GET", $url, false, $config->do_auth);

        header("content-type", $r['headers']["content-type"][0]);
        ini_set('default_mimetype', $r['headers']["content-type"][0]);
        
        if (isset($r['headers']["content-length"])){
            header("Content-length", $r['headers']["content-length"][0]);
        }

        echo $r['result'];
        die;
    }
    
    function getImageMetadata($path){
        global $config;
        
        // This is nasty from a security perspective
        // TODO: scrub domain/scheme and use our configured one
        $url = $path;
        
        $r = $this->callAPI("HEAD", $url, false, $config->do_auth);
        return $r;
    }
    
    function getSnapShot($path){
        global $config;
        $url = $config->server . "/$path";
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        header("content-type", $r['headers']["content-type"][0]);
        ini_set('default_mimetype', $r['headers']["content-type"][0]);
        
        if (isset($r['headers']["content-length"])){
            header("Content-length", $r['headers']["content-length"][0]);
        }

        echo $r['result'];
        die;        
    }
    
    function getProject($proj_path, $enforce_filters = true){
        global $config;
        
        if (!is_numeric($proj_path) && $enforce_filters && !$this->enforceProjectFilters($proj_path)){
            http_response_code(404);
            echo "Project Not found";
            die;
        }  
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path);
        
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        return json_decode($r['result']);
    }
    
    function getProjectLabels($proj_path, $enforce_filters = true){
        global $config;
        
        if (!is_numeric($proj_path) && $enforce_filters && !$this->enforceProjectFilters($proj_path)){
            http_response_code(404);
            echo "Project Not found";
            die;
        }
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/labels";
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        $labels = json_decode($r['result']);
        
        $labels_arr = array();

        while (true){
            if (!$labels){
                break;
            }
            foreach ($labels as $label) {               
                $p = array(
                    "id" => $label->id,
                    "name" => $label->name,
                    "bgcolour" => $label->color,
                    "textcolour" => $label->text_color,
                    "description" => $label->description_html
                );

                $labels_arr["label-" . $label->name] = $p;
            }
            
            if (array_key_exists("next", $r['links'])){
                $b = $this->callAPI("GET", GILSUtils::forceHTTPS($r['links']["next"]["link"]), false, $config->do_auth);
                $r = $b;
                $labels = json_decode($r['result']);
            }else{
                break;
            }
        }        
        
        return $labels_arr;
        
    }
    
    function getIssueAndComments($proj_path, $issue_ref){
    // TODO - need to add project info
        // /projects/:id/issues/:issue_iid
    
        global $config;
        
        if (!$this->enforceProjectFilters($proj_path)){
            http_response_code(404);
            echo "Issue Not found";
            die;
        }        
        
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/issues/" . $issue_ref;
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        //print_r($r);
        
        if ($r['result'] == '{"message":"404 Not found"}'){
            http_response_code(404);
            echo "Not found";
            die;
        }
        
        $issue = json_decode($r['result']);

        
        // Get label info
        $labels = $this->getProjectLabels($proj_path, false);
        
        $p = array(
            "title" => $issue->title,
            "description" => GILSUtils::replaceCommitRefsEtc(str_replace("](/uploads", "](uploads",$issue->description), $proj_path),
            "state" => $issue->state,
            "created" => $issue->created_at,
            "closed" => $issue->closed_at,
            "closed_by" => $issue->closed_by,
            "labels" => array(),
            "milestone" => $issue->milestone,
            "assignees" => $issue->assignees,
            "author" => $issue->author,
            "type" => $issue->issue_type,
            "assignee" => $issue->assignee,
            "gitlab_url" => $issue->web_url,
            "gitlab_reference" =>  $issue->references->full,
            "short_reference" =>  $issue->references->short,
        );        
        
        
        if (count($labels) > 0){
            // Add label information
            foreach($issue->labels as $label){
                $lkey = "label-".$label;
                // AFAIK This should never evaluate false, but play it safe
                if (array_key_exists($lkey, $labels)){
                    $p["labels"][] = array(
                                "text" => $labels[$lkey]['name'],
                                "id" => $labels[$lkey]['id'],
                                "description" => $labels[$lkey]['description']
                                    );
                }
            }
        }
        
        $o = array("issue" => $p, 
                    "comment_count" => 0,
                    "project" => $this->getProject($issue->project_id),
                    "related" => array()
                    );
        
        // Get related issues
        $url = $config->server . "/api/v4/projects/" . urlencode($proj_path) . "/issues/" . $issue_ref ."/links";
        $r = $this->callAPI("GET", $url, false, $config->do_auth);
        $relissues = json_decode($r['result']);
        
        foreach($relissues as $rel){
            $p = array(
                    "id" => $rel->iid,
                    "title" => $rel->title,
                    "state" => $rel->state,
                    "milestone" => $rel->milestone,
                    "gitlab_reference" => $rel->references->full
                 );
            $o["related"][] = $p;
        }
        
        // Now we need to get comments
        $url = GILSUtils::forceHTTPS($issue->_links->notes);
        $r = $this->callAPI("GET", $url."?sort=asc&order_by=created_at", false, $config->do_auth);
        //print_r($r);
        
        
        $comments= json_decode($r['result']);
        $comments_arr = array();
        $x = 0;
        while (true){
            if (!$comments){
                break;
            }
            foreach ($comments as $comment) {

                $body = GILSUtils::replaceCommitRefsEtc($comment->body, $proj_path);
                $body = str_replace("](/uploads", "](uploads", $body);
                
                $p = array(
                    "id" => $comment->id,
                    "body" => $body,
                    "author" => $comment->author,
                    "created" => $comment->created_at,
                    "update" => $comment->updated_at,
                    "system_update" => $comment->system,
                    "rowtype" => "comment",
                );
                
                if (!$comment->system){
                    $o["comment_count"]++;
                }else{
                    $p['rowtype'] = "statechange";

                    // If enhanced commit notifs are on, build an array for the template to consume
                    if ($config->enhanced_commit_notifications && str_contains($comment->body, $config->commit_mention_text)){
                        
                        $parts = explode(" ", $comment->body);
                        $commit_full_ref = array_pop($parts);
                        $commit_parts = explode("@", $commit_full_ref);
                        
                        if (count($commit_parts) == 2){
                            if (!str_contains($commit_parts[0], "/")){
                                // It's a partial reference - referencing project
                                // is in the same group as us
                                // websites/Gitlab-Issue-Listing-Script#41

                                $parts = explode("/", $proj_path);
                                $commit_parts[0] = $parts[0] . "/" . $commit_parts[0];
                            }

                            $commit_dets = $this->getCommit($commit_parts[0], $commit_parts[1], false);
                        } else {
                            // It's a relative commit
                            $commit_dets = $this->getCommit($proj_path, $commit_parts[0], false);
                        }
                        //print_r($commit_dets);
                        
                        $p["rowtype"] = "commit";
                        $p["commit_author"] = $commit_dets['commit']->author_name;
                        $p["commit_ref"] = $commit_full_ref;
                        $p["commit_date"] = $commit_dets['commit']->committed_date;
                        $p["commit_body"] = $commit_dets['commit']->message;
                        $p["additions"] = $commit_dets['commit']->stats->additions;
                        $p["deletions"] = $commit_dets['commit']->stats->deletions;
                        $p["total_changes"] = $commit_dets['commit']->stats->total;
                        $p["signature"] = $commit_dets["signature"];
                        
                    }
                }
                
                $comments_arr[] = $p;
            }
            
            if (array_key_exists("next", $r['links'])){
                $b = $this->callAPI("GET", GILSUtils::forceHTTPS($r['links']["next"]["link"]), false, $config->do_auth);
                $r = $b;
                $comments = json_decode($r['result']);
            }else{
                break;
            }
            $x++;
        }
        
        $o["comments"] = $comments_arr;
        return $o;
        
    }


    function callAPI($method, $url, $data = false, $do_auth = true)
    {
        // based on https://stackoverflow.com/a/9802854
        
        $url = trim($url);

        $headers = [];
        global $config;

        if (!$this->curl){
            $this->curl = curl_init();
        }
        
        if ($config->redis_enabled && !$this->redis){
            $this->redis = new Redis();
            
            try {
                $this->redis->connect($config->redis_host, $config->redis_port);
                
                if ($config->redis_pass){
                    $this->redis->auth($config->redis_pass);   
                }
            }
            catch (exception $e){
                // We failed to connect
                // let's not kill the application
                // The library will push warnings out, so it's not a silent failure
                $this->redis = false;
            }
        }
        
        if ($this->redis){
            $redis_key = $config->redis_prefix . ":METHOD-$method;url=$url";
            
            // See whether we have the result cached
            $cache = $this->redis->get($redis_key);
            if ($cache){
                $cache = unserialize($cache);
                return $cache;
            }
        }
        

        switch ($method)
        {
            case "POST":
                curl_setopt($this->curl, CURLOPT_POST, 1);

                if ($data)
                    curl_setopt($this->curl, CURLOPT_POSTFIELDS, $data);
                break;
            case "PUT":
                curl_setopt($this->curl, CURLOPT_PUT, 1);
                break;
                
            case "HEAD":
                curl_setopt($this->curl, CURLOPT_NOBODY, 1);
                curl_setopt($this->curl, CURLOPT_HEADER, 1);
                break;
                
            default:
                if ($data)
                    $url = sprintf("%s?%s", $url, http_build_query($data));
        }
        
        if ($do_auth){
            if ($config->access_token){
                curl_setopt($this->curl, CURLOPT_HTTPHEADER, array(
                    'Authorization: Bearer ' . $config->access_token
                ));        
            }
        }

        curl_setopt($this->curl, CURLOPT_URL, $url);
        curl_setopt($this->curl, CURLOPT_RETURNTRANSFER, 1);

        curl_setopt($this->curl, CURLOPT_HEADERFUNCTION,
        function($curl, $header) use (&$headers)
        {
            $len = strlen($header);
            $header = explode(':', $header, 2);
            if (count($header) < 2) // ignore invalid headers
            return $len;

            $headers[strtolower(trim($header[0]))][] = trim($header[1]);

            return $len;
        }
        );


        $result = curl_exec($this->curl);

        $links = array();
        if (array_key_exists("link", $headers)){
            $links = GILSUtils::parseLinkHeader($headers['link'][0]);
        }
        $res_array = array("result" => $result, "headers" => $headers, "links" => $links);
        
        if ($this->redis){
            $this->redis->set($redis_key, serialize($res_array));
            $this->redis->expire($redis_key, $config->redis_ttl);
        }
        
        return $res_array;
    }
}
